(function (root, factory) {
    root.iotController = factory(root, root._, root.hcapPromise);
}(typeof window !== 'undefined' ? window : this, function (global, _, hcapPromise) {
    var VERSION = '20181015';
    var DEBUG = false;
    var EVENT = {
        OFF: 0,
        ON: 1,
        DIMMING: 2
    };
    var voiceIntents = ["id_curtain_on", "id_curtain_off", "id_light_on", "id_light_off", "id_light_off", "id_light_brightness_max",
        "id_increase_light_brightness", "id_decrease_light_brightness", "id_open_door", "id_close_door", "id_set_mode"];
    var _initialized = false;
    var _callId = 1;
    var _subscriptions = {};
    var _callIdResults = {};
    var _thingList = {};
    var _thingValues = {};
    var _iotGroup = {
        name: "",
        aliases: [],
        scenes: []
    };
    var retry = 0;

    function log(message) {
        if (DEBUG) {
            if (_.isPlainObject(message)) message = JSON.stringify(message);
            message = '--- iotController : ' + message + ' ---';
            global.console.error(message);
        }
    }

    function redefineFunction() {
        /* Change function */
        iotController.init = RMS.init;
        iotController.requestSetComponent = RMS.setThingInfo;
        iotController.updateThingStatus = RMS.updateThingStatus;
    }

    function init() {
        logConsole("-- IoT Zigbee init");
        hcapPromise.iot.getFrameworkStatus({
            "onSuccess": function(s) {
                logConsole("-- IoT Zigbee getFrameworkStatus success");
                if (s.status == "ready") {
                    getThingList();
                } else {
                    if (retry < 5) {
                        setTimeout(function() {
                            logConsole("-- IoT retry : " + retry);
                            iotController.init();
                            ++retry;
                        },500);
                    }
                }
            }, 
            "onFailure" : function(f) {
                logConsole("-- IoT Zigbee getFrameworkStatus fail : " + f.errorMessage);
            }
       });
    }

    function getThingList() {
        logConsole("-- IoT Zigbee getThingList");
        _iotGroup = {
            name: localStorage.getItem('iotGroupName'),
            scenes: localStorage.getItem('iotScenes') ? JSON.parse(localStorage.getItem('iotScenes')) : []
        };

        // Zigbee logic
        hcapPromise.iot.getThingList()
        .then(function (s) {
            _thingList = _.chain(s.thingList)
                .filter(function (v) {
                    return !!_.get(v, 'thingMetaData.nickname');
                })
                .map(function (v) {
                    return {
                        thingId: v.thingId,
                        thingUid: v.thingUid,
                        thingType: v.thingType,
                        thingValue: {},
                        nickname: v.thingMetaData.nickname,
                        thingComponentList: _.map(v.thingComponentList, function (c) {
                            return _.pick(c, ['componentId', 'componentValue'])
                        })
                    }
                })
                .reduce(function (o, v) {
                    o[v.nickname] = v;
                    return o;
                }, {})
                .value();

            updateThingStatus();
            _initialized = true;
            retry = 0;
            logConsole("-- IoT Zigbee initialized");
        });
    }

    function getComponentByValue(component, id, value) {
        if (_.isUndefined(value)) value = "";
        if (id === EVENT.ON && !_.isEmpty(value)) {
            id = EVENT.DIMMING;
        }
        var eventIds = [];
        switch (id) {
            case EVENT.ON:
                eventIds = ['On', 'Open', 'Unlock'];
                break;
            case EVENT.OFF:
                eventIds = ['Off', 'Close', 'Lock'];
                break;
            case EVENT.DIMMING:
                eventIds = ['Dimming'];
                break;
        }
        var c = _.find(component.thingComponentList, function (v) {
            return _.includes(eventIds, v.componentId)
        });
        c.componentValue.value = value;
        if(id === EVENT.DIMMING) { c.componentValue.valueType = "Double"; }
        return c;
    }

    function buildRequestObj(request, _callback) {
        var component = _thingList[request.nickname];
        if (!_.isUndefined(component)) {
            var id = _callId++;
            _callIdResults[id] = [request.nickname, request.id, request.value];
            return _.assign({
                    thingUid: component.thingUid,
                    callId: id,
                    onSuccess: function (s) {
                        _callback(request.nickname, request.id, true);
                    },
                    onFailure: function (s) {
                        delete _thingValues[request.nickname];
                        _callback(request.nickname, request.id, false);
                    }
                },
                getComponentByValue(component, request.id, request.value)
            );
        }
    }

    function requestSetComponent(request, _callback) {
        if (!_.isFunction(_callback)) {
            _callback = _.noop;
        }

        if (_thingValues[request.nickname] && Date.now() - _thingValues[request.nickname].timestamp < 5000) {
            return Promise.resolve();
        }

        if (_thingList[request.nickname] && _thingList[request.nickname].thingType != "ThingWindowCoveringDevice") {
            _thingValues[request.nickname] = {
                request: request,
                timestamp: Date.now()
            };
        }

        var reqObj = buildRequestObj(request, _callback);
        return _.isUndefined(reqObj) ? Promise.resolve() : hcapPromise.iot.requestSetComponent(reqObj);
    }

    function requestSetComponents(requests) {
        Promise.resolve(requests)
            .mapSeries(function (request) {
                var COMMAND_DELAY = 100;
                return requestSetComponent(request).delay(COMMAND_DELAY);
            })
            .then(function (v) {
                log(v);
            });
    }

    function requestSetComponentsByScene(scene) {
        var filteredActions = _.filter(scene.actions, function (v) {
            return !!_thingList[v.thingAliasName];
        });

        requestSetComponents(_.map(filteredActions, function (v) {
            var actionValue = _.first(JSON.parse(v.thingComponentList));

            return {
                nickname: v.thingAliasName,
                id: Number(actionValue.id),
                value: actionValue.value || ""
            }
        }));
    }

    function requestScene(sceneName) {
        var scene = _.find(_iotGroup.scenes, {
            name: sceneName
        });
        if (!_.isUndefined(scene)) {
            requestSetComponentsByScene(scene);
        }
    }

    function requestEvent(eventName) {
        var scene = _.find(_iotGroup.scenes, {
            event: eventName
        });
        if (!_.isUndefined(scene)) {
            requestSetComponentsByScene(scene);
        }
    }

    function requestVoiceIntent(intentId) {
        var scene = _.find(_iotGroup.scenes, function (o) {
            return o.intents && o.intents.indexOf(intentId) > -1;
        });

        if (!_.isUndefined(scene)) {
            requestSetComponentsByScene(scene);
        }
    }

    function isVoiceintent(intentId){
        return _.indexOf(voiceIntents, intentId);
    }

    function subscribe(nickname, callback) {
        if (!_.has(_subscriptions, nickname)) {
            _subscriptions[nickname] = [];
        }
        _subscriptions[nickname].push(callback);
    }

    function publish(nickname) {
        _.forEach(_subscriptions[nickname], function (_callback) {
            _callback(_thingList[nickname]);
        });
    }

    function unSubscribe() {
        _subscriptions = {};
        return true;
    }

    function event(eventId, param) {
        if (eventId === 'iot_component_reported') {
            var isThingWindowCoveringDevice = false;
            var openStatue = 0;
            if( param.thingUid === "Zigbee:ZHA" ){
                var fromValue = _.attempt(JSON.parse.bind(null, param.fromValue.value));
                if(_.isError(fromValue)){
                    fromValue = {};
                }
                var cmd = _.get(fromValue, 'success.request.zcmd');
                if(!_.isUndefined(cmd) && cmd.indexOf("window-covering") !== -1) {
                    param.thingUid = "Zigbee:"+ fromValue.success.request.uid;
                    isThingWindowCoveringDevice = true;
                    if(cmd.indexOf("window-covering down") !== -1) {
                        openStatue = 1;  /* 1 : off, 0 : on*/
                    }
                } else { // TV model
                    var toValue = _.attempt(JSON.parse.bind(null, param.toValue.value));
                    if(!_.isError(toValue)) {
                        if (toValue.uid) {
                            if (_.find(_thingList, { thingUid: 'Zigbee:' + toValue.uid, thingType: 'ThingWindowCoveringDevice'})
                                    &&  Array.isArray(toValue.raw) && toValue.raw.length === 1 && (toValue.value === 0 || toValue.value === 100)) {
                                param.thingUid = "Zigbee:" + toValue.uid;
                                isThingWindowCoveringDevice = true;
                                openStatue = toValue.value ? 0 : 1;
                            }
                        }
                    }
                }
            }

            var nickname = _.findKey(_thingList, {
                thingUid: param.thingUid
            });
            if (!_.isUndefined(nickname)) {
                if(isThingWindowCoveringDevice){
                    _thingList[nickname].thingValue["OpenClose"] = openStatue;
                } else {
                    _thingList[nickname].thingValue[param.componentId] = param.toValue.value;
                }

                delete _thingValues[nickname];
                publish(nickname);
                log([nickname, param.componentId, param.toValue.value]);
                if(_.isPlainObject(global.logStorage)) {
                    global.logStorage.sendEvent(global.logStorage.code.IOT.CHANGED, nickname, param.toValue.value);
                }
            }
        }
    }

    function getCurrentStatus(nickname) {
        if(_.isEmpty(_thingList[nickname].thingValue)) {
            updateThingStatus(_thingList[nickname].thingUid);
        }
        return _thingList[nickname].thingValue;
    }

    function updateThingStatus(thingUid) {
        if(_.isUndefined(thingUid)) {
            _.forEach(_thingList, function(v, k){
                updateThingStatus(v.thingUid);
            });
        } else {
            hcapPromise.iot.requestSynchronizeThing({
                "thingUid": thingUid
            });
        }
    }

    /* HSP 3.5 Function */
    const RMS = {
        THING_TYPES: ['Light', 'ThingDimmableLight', 'Curtain'],
        init: function() {
            /* HSP IoT initialization
               1. Get IoT Group Name from PCN (HSP Server)
               2. Get thing information  */
            logConsole('-- HSP IoT init');
            RMS.setIoTGroupName();
            RMS.getThingList();
        },
        setIoTGroupName: function () {
            /* Get IoT group name from PCN and 
               set group name on TV (without entering pairing page) */
            logConsole('-- HSP IoT setIoTGroupName');

            requestHandler.requestAM('getIoTGroupName', '', {}, function(res) {
                if (res.status === 'success') {
                    logConsole('-- HSP IoT getIoTGroupName : ' + JSON.stringify(res.data.iotGroupName));
                    localStorage.setItem('iotGroupName', res.data.iotGroupName);
                } else {
                    logConsole('-- HSP IoT getIoTGroupName failed');
                }
            });
        },
        getThingList: function () {
            /* Get IoT thing list and status and
               save the info in legacy format */
            logConsole('-- HSP IoT getThingList');

            requestHandler.requestAM('getThingList', '', {}, function (res) {
                if (res.status == "success") {
                    _thingList = _.chain(res.data)
                        .filter(function (v) {
                            return (!!_.get(v, 'thingType') && _.includes(RMS.THING_TYPES, v.thingType))
                        })
                        .map(function (v) {
                            var _thingValue = {};
                            var _thingComponentList = _.map(v.thingComponentList, function (c) {
                                /* Update status */
                                var _id = c.componentId;
                                    _thingValue[_id] = c.componentValue.value;
                                    return _.pick(c, ['componentId', 'componentValue']);
                            });

                            return {
                                thingId: v.thingId,
                                thingUid: v.thingUid,
                                thingType: v.thingType,
                                nickname: v.thingMetaData.nickname,
                                thingValue: _thingValue,
                                thingComponentList: _thingComponentList
                            }
                        })
                        .reduce(function (o, v) {
                            o[v.nickname] = v;
                            return o;
                        }, {})
                        .value();

                    _initialized = true;
                }
            });
        },
        getThingInfo: function (thingId) {
            /* Get current thing status (checkin, guest key exception handling) */
            logConsole('-- HSP IoT getThingInfo');

            requestHandler.requestAM('getThingInfo', '', {
                "thingId": thingId
            }, function (res) {
                if (res.status === 'success') {
                    RMS.updateThingStatus(res.data);
                } else {
                    logConsole('-- HSP IoT getThingInfo failed');
                }
            });
        },
        setThingInfo: function (req, _cb) {
            /* Change thing status */
            logConsole('-- HSP IoT setThingInfo');

            if (_.isUndefined(req.value)) req.value = "";
            var _component = _thingList[req.nickname];
            var _thingType = _component.thingType;
            var eventId = '';

            if (_thingType === 'Curtain') {
                eventId = 'OpenClose';
            } else {
                if (req.id === EVENT.ON && !_.isEmpty(req.value)) {
                    req.id = EVENT.DIMMING;
                }
                eventId = (req.id === EVENT.DIMMING) ? 'Dimming' : 'OnOff';
            }

            switch (req.id) {
                case EVENT.ON:
                    req.value = String(EVENT.ON);
                    break;
                case EVENT.OFF:
                    req.value = String(EVENT.OFF);
                    break;
            }
            var _componentList = _.find(_component.thingComponentList, function(v) {
                return v.componentId === eventId;
            });
            _componentList.componentValue.value = req.value;
            
            requestHandler.requestAM('setThingInfo', '', {
                "thingId": _component.thingId,
                "thingComponentList": JSON.stringify(_componentList)
            }, function(res) {
                if (res.status === 'success') {
                    RMS.getThingInfo(_component.thingId);
                }
            });
        },
        updateThingStatus: function (data) {
            /* Update _thingList and publish */
            logConsole('-- HSP IoT updateThingStatus');

            if(!_.isUndefined(data) && _.includes(RMS.THING_TYPES, data.thingType)) {
                var _nickname = data.thingMetaData.nickname;
                _.each(data.thingComponentList, function(v) {
                    var _componentId = v.componentId;
                    _thingList[_nickname].thingValue[_componentId] = v.componentValue.value;
                });
                publish(_nickname);
            }
        }
    };

    return {
        version: VERSION,
        get initialized() {
            return _initialized
        },
        get iotGroup() {
            return _iotGroup
        },
        get thingList() {
            return _thingList
        },
        init: init,
        requestSetComponents: requestSetComponents,
        requestSetComponent: requestSetComponent,
        requestScene: requestScene,
        requestEvent: requestEvent,
        requestVoiceIntent : requestVoiceIntent,
        isVoiceIntent : isVoiceintent,
        unSubscribe: unSubscribe,
        subscribe: subscribe,
        publish: publish,
        event: event,
        getCurrentStatus: getCurrentStatus,
        redefineFunction: redefineFunction,
        updateThingStatus: updateThingStatus
    };
}));